# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import

# def get_package_data():
#     """Get Package Data for utilipy.scripts"""
#     return {'utilipy.scripts': ['data/*']}
# # /def
